#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 18 13:53:36 2024

@author: mike
"""
import pandas as pd
import os
import pathlib
import pytest

from npsfm import NPSFM

########################################################3
### Parameters

script_path = pathlib.Path(os.path.realpath(os.path.dirname(__file__)))
data_path = script_path.parent.parent.joinpath('data')

nzsegment = 3076139
parameter = 'Nitrate'
feature = 'river'
version = 'v202401'
hopeful_band = 'A'

########################################################
### Test

ts_data = pd.read_csv(script_path.joinpath('test_data1.csv.zip'), index_col=0, parse_dates=True)['value']


self = NPSFM(data_path, download_files=True)


def test_add_limits():
    limits = self.add_limits(feature, parameter, nzsegment)

    assert len(limits) == 4


limits = self.add_limits(feature, parameter, nzsegment)


def test_add_stats():
    stats = self.add_stats(ts_data)

    assert len(stats) == 2


stats = self.add_stats(ts_data)


def test_calc_band():
    attr_band = self.calc_band(include_stats=['median'])
    attr_band = self.calc_band(include_stats=None)

    assert attr_band == 'C'


def test_calc_improvement_to_band():
    improve_ratio1 = self.calc_improvement_to_band(hopeful_band, include_stats=['median'])
    improve_ratio2 = self.calc_improvement_to_bottom_line(include_stats=['median'])

    improve_ratio1 = self.calc_improvement_to_band(hopeful_band)
    improve_ratio2 = self.calc_improvement_to_bottom_line()

    assert (len(improve_ratio1) == 2) and len(improve_ratio2) == 2


































































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  5 14:40:44 2023

@author: mike
"""
import io
import os
import pathlib
import pandas as pd
import geopandas as gpd
# import nzrec
import booklet

import params

pd.options.display.max_columns = 10


#######################################################
### Parameters

# data_path = pathlib.Path(os.path.join(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0], 'data'))

way_id = 3133749

nzrec_data_path = '/home/mike/git/nzrec/data'

## Extra data
lake_poly_gpkg_path = params.data_path.joinpath('lake_polygons_fenz.gpkg')

## Output
# agg_conc_csv = 'wairarapa_stream_data.csv'
# agg_conc_feather = 'river_data.feather'

lake_tags_blt_path = params.data_path.joinpath('lake_tags.blt')



#####################################################
### Processing

lakes_poly0 = gpd.read_file(lake_poly_gpkg_path)
lakes_poly0 = lakes_poly0[lakes_poly0.LFENZID > 0].copy()
lakes_poly0['LFENZID'] = lakes_poly0['LFENZID'].astype('int32')
lakes_poly1 = lakes_poly0.drop('geometry', axis=1).drop_duplicates('LFENZID').set_index('LFENZID')

lake_tags = {}
for lake_id, tags in lakes_poly1.iterrows():
    if tags['MaxDepth'] <= 7.5:
        stratified = 'polymictic'
    else:
        stratified = 'stratified'

    tags_dict = tags.to_dict()
    tags_dict['stratified'] = stratified

    lake_tags[lake_id] = tags_dict

### Save results
with booklet.open(lake_tags_blt_path, 'n', key_serializer='uint4', value_serializer='msgpack', n_buckets=10007) as f:
    f.update(lake_tags)





































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 13 15:00:43 2023

@author: mike
"""
import os
import pathlib
import pandas as pd
import numpy as np
from copy import copy
import booklet

from . import v202401, utils
# import v202401, utils


pd.options.display.max_columns = 10

######################################################
### Parameters

# features = set([limits[0] for limits in v202401.parameter_limits_dict])
# indicators = set([limits[1] for limits in v202401.parameter_limits_dict])



#######################################################
### Functions







###################################################
### Classes


# class ImportData:
#     """

#     """
#     def __init__(self, ts_data=None, indicator=None, nzsegment=None):
#         """

#         """



#     def add_data(self, ts_data, indicator, feature):
#         """

#         """
#         ## Checks
#         if not isinstance(ts_data, pd.Series):
#             raise TypeError('ts_data must be a pandas Series with a datetime index.')
#         if not isinstance(ts_data.index, pd.DatetimeIndex):
#             raise TypeError('ts_data must be a pandas Series with a datetime index.')
#         if indicator not in indicators:
#             raise ValueError(f'indicator must be one of {indicators}')
#         if feature not in features:
#             raise ValueError(f'feature must be one of {features}')




#     def add_nzsegment(self, nzsegment):
#         """

#         """





class NPSFM:
    """

    """
    def __init__(self, package_data_path, download_files=False, only_missing=True):
        """

        """
        ## File check
        missing_files = utils.check_files(package_data_path)
        if missing_files:
            if download_files:
                utils.download_files(package_data_path, only_missing=only_missing)
            else:
                raise ValueError('There are missing data files. Files would have been downloaded if download_files=True (but it is not). Othersize if this was a mistake, please check/change the data_path: {}'.format(', '.join([os.path.split(f)[-1] for f in missing_files])))

        self.data_path = pathlib.Path(package_data_path)

        ## Run checks to see what parameters are available to calc grades


    def add_limits(self, feature, parameter, nzsegment=None, lfenzid=None):
        """

        """
        ## Checks
        feature_parameter = (feature, parameter)
        if feature_parameter not in v202401.parameter_limits_dict:
            raise ValueError(f'The combo of {feature_parameter} must be one of {list(v202401.parameter_limits_dict.keys())}')
        if isinstance(nzsegment, int):
            file_path = self.data_path.joinpath('rec_tags.blt')
            tag_id = nzsegment
        elif isinstance(lfenzid, int):
            file_path = self.data_path.joinpath('lake_tags.blt')
            tag_id = lfenzid
        else:
            tag_id = None
            tags = None
        if tag_id:
            with booklet.open(file_path) as f:
                if tag_id not in f:
                    raise ValueError(f'{tag_id} not a valid id.')
                else:
                    tags = f[tag_id]

        ## Remove old ts_data and stats if they exists
        if hasattr(self, 'ts_data'):
            delattr(self, 'ts_data')
            delattr(self, 'stats')

        ## Determine the associated limits for the site and parameter
        limits = utils.get_limits(feature_parameter, tags)

        bl_limit_band = v202401.bottom_line_limits[feature_parameter]
        bl_limit = limits[bl_limit_band]

        ## Save data
        self.parameter = parameter
        self.feature = feature
        self.feature_parameter = feature_parameter
        # self.nzsegment = nzsegment
        self.class_tags = tags
        self.bottom_line_limit_band = bl_limit_band
        self.bottom_line_limit = bl_limit
        self.limits = limits

        return limits


    def add_stats(self, ts_data):
        """

        """
        ## Checks
        if not hasattr(self, 'limits'):
            raise ValueError('You must run the add_limits method first.')
        if not isinstance(ts_data, pd.Series):
            raise TypeError('ts_data must be a pandas Series with a datetime index.')
        if not isinstance(ts_data.index, pd.DatetimeIndex):
            raise TypeError('ts_data must be a pandas Series with a datetime index.')

        ## Determine the stats
        self.stats = utils.calc_stats(ts_data, self.limits)

        return self.stats


    def calc_band(self, include_stats=None):
        """

        """
        result = utils.calc_band_from_limit(self.stats, self.limits, include_stats)

        return result


    def calc_improvement_to_band(self, band, include_stats=None):
        """

        """
        if band not in self.limits:
            raise ValueError(f'{band} not in the available bands: {list(self.limits.keys())}')

        results = utils.calc_improvement_to_band(self.stats, self.limits, band, include_stats)

        return results


    def calc_improvement_to_bottom_line(self, include_stats=None):
        """

        """
        results = utils.calc_improvement_to_band(self.stats, self.limits, self.bottom_line_limit_band, include_stats)

        return results




####################################################
### Testing


























































# -*- coding: utf-8 -*-
"""
Utility functions.
"""
import os
import io
import numpy as np
import pathlib
import smart_open
import smart_open.http as so_http
from time import sleep
import concurrent.futures
# import importlib
from copy import copy

from . import v202401
# import v202401

so_http.DEFAULT_BUFFER_SIZE = 524288

#########################################
### parameters

file_dict = {
             'rec_tags.blt': 'https://b2.tethys-ts.xyz/file/nz-mfe/nps-fm-2020/rec_tags.blt',
             'lake_tags.blt': 'https://b2.tethys-ts.xyz/file/nz-mfe/nps-fm-2020/lake_tags.blt',
             }


########################################
### Functions


def calc_stat(ts_data, stat, percentile_method='hazen'):
    """

    """
    if stat == 'median':
        value = ts_data.median()
    elif stat == 'max':
        value = ts_data.max()
    elif 'Q' in stat:
        percentile = int(stat[1:])
        value = np.percentile(ts_data, percentile, method=percentile_method)
    elif 'G' in stat:
        conc = int(stat[1:])
        value = (ts_data > conc).sum()

    return value


def calc_stats(ts_data, limits):
    """

    """
    stats = {stat: calc_stat(ts_data, stat) for stat in limits['A']}

    return stats


def calc_band_from_limit(stats, limits, include_stats=None):
    """

    """
    new_band = None

    for band, limit in reversed(limits.items()):
        bool_list = []
        if isinstance(include_stats, (list, str)):
            if isinstance(include_stats, str):
                include_stats = [include_stats]
            for stat in include_stats:
                if stat in limit:
                    min1, max1 = limit[stat]
                    bool0 = (stats[stat] > min1) & (stats[stat] <= max1)
                    bool_list.append(bool0)
                else:
                    raise ValueError(f'{stat} not in limits.')
        elif include_stats is not None:
            raise TypeError('include_stats must be either a str or a list of str.')
        else:
            for stat_name, minmax in limit.items():
                min1, max1 = minmax
                bool0 = (stats[stat_name] > min1) & (stats[stat_name] <= max1)
                bool_list.append(bool0)

        if all(bool_list):
            new_band = band

    return new_band


def get_limits(feature_parameter, tags):
    """

    """
    # nps_mod = importlib.import_module(version)
    nps_mod = v202401 # One day make this flexible...

    ## Get the limits
    limits = nps_mod.parameter_limits_dict[feature_parameter]

    ## Assign appropriate limits if it's a complicated limit...
    if feature_parameter in nps_mod.parameter_special_cols_dict:
        if tags is None:
            raise ValueError('tags must be assigned if there are special classes in the attribute.')
        tag_name = nps_mod.parameter_special_cols_dict[feature_parameter]
        tag = tags[tag_name]
        old_limits = copy(limits)
        limits = {}
        for band, lm in old_limits.items():
            limit = lm[tag]
            limits[band] = limit

    return limits


def calc_improvement_to_band(stats, limits, band, include_stats=None):
    """

    """
    current_band = calc_band_from_limit(stats, limits, include_stats)

    if band >= current_band:
        results = {stat: 0 for stat in stats}
    else:
        band_limits = limits[band]
        results = {}
        if isinstance(include_stats, (list, str)):
            if isinstance(include_stats, str):
                include_stats = [include_stats]
            for stat in include_stats:
                limit = band_limits[stat]
                current_val = stats[stat]
                if limit[0] == -1:
                    ratio = round(1 - limit[1]/current_val, 4)
                else:
                    ratio = round(limit[0]/current_val - 1, 4)
    
                results[stat] = ratio
        elif include_stats is not None:
            raise TypeError('include_stats must be either a str or a list of str.')
        else:
            for stat, limit in band_limits.items():
                current_val = stats[stat]
                if limit[0] == -1:
                    ratio = round(1 - limit[1]/current_val, 4)
                else:
                    ratio = round(limit[0]/current_val - 1, 4)

                results[stat] = ratio

    return results


def url_to_file(url, file_path, chunk_size: int=524288, retries=3):
    """
    General function to get an object from an S3 bucket. One of s3, connection_config, or public_url must be used.

    Parameters
    ----------
    url: http str
        The http url to the file.
    chunk_size: int
        The amount of bytes to download as once.

    Returns
    -------
    file object
        file object of the S3 object.
    """
    transport_params = {'buffer_size': chunk_size, 'timeout': 120}

    ## Get the object
    counter = retries
    while True:
        try:
            file_obj = smart_open.open(url, 'rb', transport_params=transport_params, compression='disable')
            file_path1 = pathlib.Path(file_path)
            file_path1.parent.mkdir(parents=True, exist_ok=True)

            with open(file_path1, 'wb') as f:
                chunk = file_obj.read(chunk_size)
                while chunk:
                    f.write(chunk)
                    chunk = file_obj.read(chunk_size)
            break
        except Exception as err:
            counter = counter - 1
            if counter > 0:
                sleep(3)
            else:
                print('smart_open could not open url with the following error:')
                print(err)
                file_obj = None
                break

    return file_obj


def check_files(data_path):
    """

    """
    data_path1 = pathlib.Path(data_path)

    files = [f.name for f in data_path1.glob('*') if f.name in file_dict]
    missing_files = []
    for f in file_dict:
        if f not in files:
            missing_files.append(file_dict[f])

    return missing_files


def download_files(data_path, only_missing=True):
    """

    """
    if only_missing:
        urls = check_files(data_path)
    else:
        urls = list(file_dict.values())

    print('Downloading: {}'.format(', '.join([os.path.split(url)[-1] for url in urls])))
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        futures = []
        for url in urls:
            file_name = os.path.split(url)[-1]
            new_path = os.path.join(data_path, file_name)
            f = executor.submit(url_to_file, url, new_path)
            futures.append(f)
        _ = concurrent.futures.wait(futures)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 14 10:28:31 2024

@author: mike
"""
import os
import pathlib

###################################################
### REC parameters

script_path = pathlib.Path(os.path.realpath(os.path.dirname(__file__)))
data_path = script_path.parent.joinpath('data')

if not data_path.is_dir():
    raise ValueError(f'{data_path} does not exist')




















































































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  5 14:40:44 2023

@author: mike
"""
import io
import os
import pathlib
import pandas as pd
import geopandas as gpd
import nzrec
import booklet

import params

pd.options.display.max_columns = 10


#######################################################
### Parameters

# data_path = pathlib.Path(os.path.join(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0], 'data'))

way_id = 3133749

nzrec_data_path = '/home/mike/git/nzrec/data'

## Extra data
sed_csv = 'sediment-classes-for-rec24-nzsegments.csv.zip'

## Output
# agg_conc_csv = 'wairarapa_stream_data.csv'
# agg_conc_feather = 'river_data.feather'

rec_classes_blt_path = params.data_path.joinpath('rec_tags.blt')



#####################################################
### Processing

rec_tags = {}

w0 = nzrec.Water(nzrec_data_path)


# tags = w0._way_tag[way_id]

for way_id, tags in w0._way_tag.items():

    ## Periphyton classes
    climate = tags['Climate class']
    geology = tags['Geology class']

    if (climate in ('WD', 'CD')) and (geology in ('SS', 'VA', 'VB')):
        peri_class = 2
    else:
        peri_class = 1

    rec_tags[way_id] = {'peri_class': peri_class}

    ## SS class
    ss_class = tags['Suspended_4_class']
    if (ss_class is None):
        ss_class = 1
    rec_tags[way_id].update({'ss_class': ss_class})

    ## deposited sediment class
    ds_class = tags['Deposited_4_class']
    if (ds_class is None) or (ds_class == 'naturally soft-bottomed'):
        ds_class = 0
    rec_tags[way_id].update({'ds_class': ds_class})

### Save results
# data_path.mkdir(parents=True, exist_ok=True)

with booklet.open(rec_classes_blt_path, 'n', key_serializer='uint4', value_serializer='msgpack', n_buckets=6000011) as f:
    f.update(rec_tags)

w0.close()

































